/*
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 1, or (at your option)
** any later version.

** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.

** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
 $Author: nobunaga $
 $Revision: 1.12 $
 $Date: 2007/12/18 08:56:48 $
*/

/*
* Author :
* Boris Mikhailenko <stg34@stargazer.dp.ua>
* Andrey Rakhmanov <andrey_rakhmanov@yahoo.com> -   .
*/

//---------------------------------------------------------------------------

#include <iconv.h>

#ifdef WIN32
    #include <windows.h>
    #include <winbase.h>
    #include <winnt.h>
#else
    #include <fcntl.h>
    #include <sys/types.h>
    #include <sys/socket.h>
    #include <netdb.h>
    #include <netinet/in.h>
    #include <arpa/inet.h>
    #include <sys/time.h>
    #include <unistd.h>
#endif

#include <cassert>
#include <cerrno>
#include <cstring>
#include <iostream>
#include <vector>

#include <boost/detail/endian.hpp>

#include "common.h"
#include "ia_auth_c.h"

#define IA_ID "00100"
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
namespace {

#ifndef WIN32
uint32_t GetTickCount()
{
    struct timeval tv;
    gettimeofday(&tv, NULL);
    return tv.tv_sec * 1000 + tv.tv_usec / 1000;
}
#endif

bool IconvString(const std::string & source,
                 std::string & dest,
                 const std::string & from,
                 const std::string & to)
{
    if (source.empty())
        return true;

    size_t inBytesLeft = source.length();
    size_t outBytesLeft = source.length() * 2;

    std::vector<char> inBuf(source.begin(), source.end());
    std::vector<char> outBuf(outBytesLeft);

#if defined(WIN32)
    const char * srcPos = &inBuf.front();
#else
    char * srcPos = &inBuf.front();
#endif
    char * dstPos = &outBuf.front();

    iconv_t handle = iconv_open(to.c_str(),
                                from.c_str());

    if (handle == iconv_t(-1)) {
        if (errno == EINVAL) {
            std::cerr << "IconvString(): iconv from " << from << " to " << to << " failed" << std::endl;
            return false;
        } else {
            printfd(__FILE__, "IconvString(): iconv_open error\n");
        }
        return false;
    }

    size_t res = iconv(handle,
                       &srcPos, &inBytesLeft,
                       &dstPos, &outBytesLeft);

    if (res == size_t(-1)) {
        printfd(__FILE__, "IconvString(): '%s'\n", strerror(errno));

        iconv_close(handle);
        return false;
    }

    dest.assign(&outBuf.front(), dstPos);

    iconv_close(handle);

    return true;
}

}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
IA_CLIENT_PROT::IA_CLIENT_PROT(const std::string & serverName,
                               uint16_t port,
                               const std::string & login,
                               const std::string & password)
    : m_phase(1),
      m_phaseTime(0),
      m_running(false),
      m_login(login),
      m_firstConnect(true),
      m_reconnect(false),
      m_sockr(-1),
      m_userTimeout(60),
      m_aliveTimeout(5),
      m_rnd(0)
{
#ifdef DEBUG
    std::cerr << "IA_CLIENT_PROT(server: '" << serverName << "', port: " << port << ", login: '" << login << "', password: '" << password << "')" << std::endl;
#endif
    memset(&m_loadStat, 0, sizeof(m_loadStat));
    memset(&m_selectedDirs, 0xff, sizeof(m_selectedDirs));

    m_packetTypes["CONN_SYN"] = CONN_SYN_N;
    m_packetTypes["CONN_SYN_ACK"] = CONN_SYN_ACK_N;
    m_packetTypes["CONN_ACK"] = CONN_ACK_N;
    m_packetTypes["ALIVE_SYN"] = ALIVE_SYN_N;
    m_packetTypes["ALIVE_ACK"] = ALIVE_ACK_N;
    m_packetTypes["DISCONN_SYN"] = DISCONN_SYN_N;
    m_packetTypes["DISCONN_SYN_ACK"] = DISCONN_SYN_ACK_N;
    m_packetTypes["DISCONN_ACK"] = DISCONN_ACK_N;
    m_packetTypes["FIN"] = FIN_N;
    m_packetTypes["ERR"] = ERROR_N;
    m_packetTypes["INFO"] = INFO_N;
    m_packetTypes["INFO_7"] = INFO_7_N;
    m_packetTypes["INFO_8"] = INFO_8_N;

    EnDecodeInit("pr7Hhen", 8, &m_ctxHdr);
    EnDecodeInit(password.c_str(), password.length(), &m_ctxPass);

    if (m_prepareNet(serverName, port) == -1) {
        throw std::runtime_error(m_strError);
    }
}
//---------------------------------------------------------------------------
int IA_CLIENT_PROT::m_prepareNet(const std::string & serverName, uint16_t port)
{
    m_servAddr.sin_family = AF_INET;
    m_servAddr.sin_port = htons(port);
    m_servAddr.sin_addr.s_addr = inet_addr(serverName.c_str());

    if (m_servAddr.sin_addr.s_addr == INADDR_NONE) {
        struct hostent * phe = gethostbyname(serverName.c_str());
        if (phe) {
            m_servAddr.sin_addr.s_addr = *(reinterpret_cast<unsigned long *>(phe->h_addr_list[0]));
        } else {
            m_strError = std::string("Unknown host ") + "\'" + serverName + "\'";
            // No callbacks at here
            return -1;
        }
    }

    m_sockr = socket(AF_INET, SOCK_DGRAM, 0);

    if (m_sockr < 0) {
        m_strError = strerror(errno);
        // No callbacks at here
        return -1;
    }

    return 0;
}
//---------------------------------------------------------------------------
IA_CLIENT_PROT::~IA_CLIENT_PROT()
{
#ifdef DEBUG
    std::cerr << "~IA_CLIENT_PROT" << std::endl;
#endif
    stop();
#ifdef WIN32
    closesocket(m_sockr);
#else
    close(m_sockr);
#endif
}
//---------------------------------------------------------------------------
void IA_CLIENT_PROT::start()
{
    {
        boost::lock_guard<boost::mutex> lock(m_mutex);
        if (m_running)
            return;
    }

    m_running = true;

    m_thread = boost::thread(boost::bind(&IA_CLIENT_PROT::m_run, this));
}
//---------------------------------------------------------------------------
void IA_CLIENT_PROT::stop()
{
    boost::unique_lock<boost::mutex> lock(m_mutex);
    if (!m_running)
        return;
    lock.unlock();

    disconnect();

    waitDisconnect(boost::posix_time::time_duration(0, 0, 1, 0)); // 1 sec

    lock.lock();
    m_running = false;
    lock.unlock();

    m_thread.join();
}
//---------------------------------------------------------------------------
void IA_CLIENT_PROT::m_run()
{
    while (true) {
        {
            boost::lock_guard<boost::mutex> lock(m_mutex);
            if (!m_running)
                break;
        }
        m_netRecv();
        m_checkTimeout();
    }
}
//---------------------------------------------------------------------------
void IA_CLIENT_PROT::connect()
{
    boost::unique_lock<boost::mutex> lock(m_mutex);
    if (m_phase == 2 ||
        m_phase == 3) {
        return;
    }
    lock.unlock();

    m_netSend(CONN_SYN_N);

    lock.lock();
    m_phase = 2;
    m_phaseTime = GetTickCount();
    lock.unlock();

    m_cond.notify_one();
}
//---------------------------------------------------------------------------
void IA_CLIENT_PROT::disconnect()
{
    boost::unique_lock<boost::mutex> lock(m_mutex);
    if (m_phase != 2 &&
        m_phase != 3) {
        return;
    }
    lock.unlock();

    m_netSend(DISCONN_SYN_N);

    lock.lock();
    m_phase = 4;
    m_phaseTime = GetTickCount();
    m_firstConnect = true;
    lock.unlock();

    m_cond.notify_one();
}
//---------------------------------------------------------------------------
bool IA_CLIENT_PROT::waitConnect(const boost::posix_time::time_duration & timeout)
{
    boost::unique_lock<boost::mutex> lock(m_mutex);
    while(m_phase != 2 &&
          m_phase != 3) {
        if (!m_cond.timed_wait(lock, timeout))
            return false;
        if (m_phase != 2 &&
            m_phase != 3)
            return false;
    }
    return true;
}
//---------------------------------------------------------------------------
bool IA_CLIENT_PROT::waitDisconnect(const boost::posix_time::time_duration & timeout)
{
    boost::unique_lock<boost::mutex> lock(m_mutex);
    while(m_phase == 2 ||
          m_phase == 3) {
        if (!m_cond.timed_wait(lock, timeout))
            return false;
        if (m_phase == 2 ||
            m_phase == 3)
            return false;
    }
    return true;
}
//---------------------------------------------------------------------------
void IA_CLIENT_PROT::setEnabledDirs(const bool * selectedDirs)
{
    boost::lock_guard<boost::mutex> lock(m_mutex);
    memcpy(m_selectedDirs, selectedDirs, sizeof(bool) * DIR_NUM);
}
//---------------------------------------------------------------------------
bool IA_CLIENT_PROT::isAuthorized() const
{
    boost::lock_guard<boost::mutex> lock(m_mutex);
    return m_phase == 3 || m_phase == 4;
}
//---------------------------------------------------------------------------
void IA_CLIENT_PROT::getStat(LOADSTAT * ls)
{
    boost::lock_guard<boost::mutex> lock(m_mutex);
    memcpy(ls, &m_loadStat, sizeof(m_loadStat));
}
//---------------------------------------------------------------------------
void IA_CLIENT_PROT::m_netRecv()
{
    char buffer[2048];
    int res = m_recv(buffer, sizeof(buffer));
    if (res < 0) {
        if (!m_errorCb.empty())
            m_errorCb(m_strError, m_codeError);
        return;
    }
    if (res == 0)
        return;

    switch (m_determinatePacketType(buffer + 12)) {
        case CONN_SYN_ACK_N:
            m_process_CONN_SYN_ACK_8(buffer);
            break;
        case ALIVE_SYN_N:
            m_process_ALIVE_SYN_8(buffer);
            break;
        case DISCONN_SYN_ACK_N:
            m_process_DISCONN_SYN_ACK_8(buffer);
            break;
        case FIN_N:
            m_process_FIN_8(buffer);
            break;
        case INFO_8_N:
            m_process_INFO_8(buffer);
            break;
        case ERROR_N:
            m_process_ERROR(buffer);
            break;
        default:
            return;
    }
    return;
}
//---------------------------------------------------------------------------
void IA_CLIENT_PROT::m_netSend(int n)
{
    char buffer[2048];
    size_t msgLen = 0;

    memset(buffer, 0, 2048);

    switch (n) {
        case CONN_SYN_N:
            msgLen = m_prepare_CONN_SYN_8(buffer, sizeof(buffer));
            break;
        case CONN_ACK_N:
            msgLen = m_prepare_CONN_ACK_8(buffer, sizeof(buffer));
            break;
        case ALIVE_ACK_N:
            msgLen = m_prepare_ALIVE_ACK_8(buffer, sizeof(buffer));
            break;
        case DISCONN_SYN_N:
            msgLen = m_prepare_DISCONN_SYN_8(buffer, sizeof(buffer));
            break;
        case DISCONN_ACK_N:
            msgLen = m_prepare_DISCONN_ACK_8(buffer, sizeof(buffer));
            break;
        default:
            return;
    }

    m_fillHdr8(buffer);
    if (m_send(buffer, msgLen) == -1) {
        if (!m_errorCb.empty())
            m_errorCb(m_strError, m_codeError);
    }
    return;
}
//---------------------------------------------------------------------------
int IA_CLIENT_PROT::m_recv(char * buffer, size_t len)
{
    fd_set rfds;
    struct timeval tv;

    FD_ZERO(&rfds);
    FD_SET(m_sockr, &rfds);
    tv.tv_sec = 0;
    tv.tv_usec = 500000;

    int ret = select(m_sockr + 1, &rfds, NULL, NULL, &tv);

    if (ret == -1) {
        // Error
        boost::lock_guard<boost::mutex> lock(m_mutex);
        m_strError = strerror(errno);
        m_codeError = IA_SELECT_ERROR;
        return -1;
    } else if (ret == 0) {
        // Timeout
        return 0;
    }

    struct sockaddr_in addr;
#ifdef WIN32
    int fromLen = sizeof(addr);
#else
    socklen_t fromLen = sizeof(addr);
#endif

    if (recvfrom(m_sockr, buffer, len, 0, reinterpret_cast<struct sockaddr *>(&addr), &fromLen) == -1) {
        boost::lock_guard<boost::mutex> lock(m_mutex);
        m_strError = strerror(errno);
        m_codeError = IA_RECV_ERROR;
        return -1;
    }

    if (strncmp(buffer + 4 + sizeof(HDR_8), "ERR", sizeof("ERR"))) {
        for (size_t i = 0; i < len / 8; i++)
            DecodeString(buffer + i * 8, buffer + i * 8, &m_ctxPass);
    }

    return 1;
}
//---------------------------------------------------------------------------
int IA_CLIENT_PROT::m_send(char * buffer, size_t len)
{
    size_t db = sizeof(HDR_8);
    for (size_t i = 0; i < IA_LOGIN_LEN / 8; i++) {
        EncodeString(buffer + db + i * 8, buffer + db + i * 8, &m_ctxHdr);
    }

    db += IA_LOGIN_LEN;
    size_t encLen = (len - sizeof(HDR_8) - IA_LOGIN_LEN)/8;
    for (size_t i = 0; i < encLen; i++) {
        EncodeString(buffer + db, buffer + db, &m_ctxPass);
        db += 8;
    }

    if (sendto(m_sockr, buffer, len, 0, reinterpret_cast<struct sockaddr *>(&m_servAddr), sizeof(m_servAddr)) == -1) {
        boost::lock_guard<boost::mutex> lock(m_mutex);
        m_strError = strerror(errno);
        m_codeError = IA_SEND_ERROR;
        return -1;
    }
    return 0;
}
//---------------------------------------------------------------------------
void IA_CLIENT_PROT::m_checkTimeout()
{
    bool connect = false;
    bool timeout = false;
    boost::unique_lock<boost::mutex> lock(m_mutex);
    switch (m_phase) {
        case 1:
            if (m_reconnect && !m_firstConnect) {
                connect = true;
                m_phase = 2;
                m_phaseTime = GetTickCount();
            }
            break;
        case 2:
            if ((GetTickCount() - m_phaseTime) / 1000 > m_aliveTimeout) {
                m_phase = 1;
                m_phaseTime = GetTickCount();
                timeout = true;
            }
            break;
        case 3:
            if ((GetTickCount() - m_phaseTime) / 1000 > m_userTimeout) {
                m_phase = 1;
                m_phaseTime = GetTickCount();
                m_firstConnect = false;
                timeout = true;
            }
            break;
        case 4:
            if ((GetTickCount() - m_phaseTime) / 1000 > m_aliveTimeout) {
                m_phase = 1;
                m_phaseTime = GetTickCount();
                timeout = true;
            }
            break;
        case 5:
            if ((GetTickCount() - m_phaseTime) / 1000 > m_aliveTimeout) {
                m_phase = 1;
                m_phaseTime = GetTickCount();
                timeout = true;
            }
            break;
    }
    lock.unlock();
    if (timeout) {
        m_cond.notify_one();
        if (!m_errorCb.empty())
            m_errorCb("Packet timeout", IA_TIMEOUT);
        if (!m_statusCb.empty())
            m_statusCb(false);
    }
    if (connect) {
        m_netSend(CONN_SYN_N);
        m_cond.notify_one();
    }
}
//---------------------------------------------------------------------------
void IA_CLIENT_PROT::m_fillHdr8(char * buffer)
{
    strncpy(buffer, IA_ID, 6);
    buffer[IA_MAGIC_LEN] = 0;
    buffer[IA_MAGIC_LEN + 1] = IA_PROTO_VER;
    strncpy(buffer + sizeof(HDR_8), m_login.c_str(), IA_LOGIN_LEN);
}
//---------------------------------------------------------------------------
std::string IA_CLIENT_PROT::getError() const
{
    boost::lock_guard<boost::mutex> lock(m_mutex);
    return m_strError;
}
//---------------------------------------------------------------------------
int IA_CLIENT_PROT::getErrorCode() const
{
    boost::lock_guard<boost::mutex> lock(m_mutex);
    return m_codeError;
}
//---------------------------------------------------------------------------
void IA_CLIENT_PROT::m_process_CONN_SYN_ACK_8(const void * buffer)
{
    const CONN_SYN_ACK_8 * connSynAck8 = static_cast<const CONN_SYN_ACK_8 *>(buffer);
    {
        boost::lock_guard<boost::mutex> lock(m_mutex);
        m_rnd = connSynAck8->rnd;
        m_userTimeout = connSynAck8->userTimeOut;
        m_aliveTimeout = connSynAck8->aliveDelay;
    }

#ifdef DEBUG
    std::cerr << ">> CONN_SYN_ACK (rnd: " << m_rnd << ", userTimeout: " << m_userTimeout << ", aliveTimeout: " << m_aliveTimeout <<  ")" << std::endl;
#endif

    std::vector<std::string> dirNames;
    for (size_t i = 0; i < DIR_NUM; i++) {
        std::string source(reinterpret_cast<const char *>(connSynAck8->dirName[i]));
        std::string dest;
        if (IconvString(source, dest, "KOI8-RU", "UTF8")) {
            dirNames.push_back(dest);
        } else {
#ifdef DEBUG
        std::cerr << "Failed to convert '" << source << "' from KOI8-RU to UTF8" << std::endl;
#endif
        }
    }

    if (!m_dirNamesCb.empty())
        m_dirNamesCb(dirNames);

    m_netSend(CONN_ACK_N);

    {
        boost::lock_guard<boost::mutex> lock(m_mutex);
        m_phase = 3;
        m_phaseTime = GetTickCount();
    }
    m_cond.notify_one();
}
//---------------------------------------------------------------------------
void IA_CLIENT_PROT::m_process_ALIVE_SYN_8(const void * buffer)
{
    const ALIVE_SYN_8 * aliveSyn8 = static_cast<const ALIVE_SYN_8 *>(buffer);
    {
        boost::lock_guard<boost::mutex> lock(m_mutex);
        m_rnd = aliveSyn8->rnd;
        memcpy(&m_loadStat, aliveSyn8->mu, sizeof(m_loadStat));
    }

#ifdef DEBUG
    std::cerr << ">> ALIVE_SYN (rnd: " << m_rnd << ")" << std::endl;
#endif

    if (!m_statCb.empty())
        m_statCb(m_loadStat);

    if (!m_statusCb.empty())
        m_statusCb(true);

    m_netSend(ALIVE_ACK_N);

    boost::lock_guard<boost::mutex> lock(m_mutex);
    m_phaseTime = GetTickCount();
}
//---------------------------------------------------------------------------
void IA_CLIENT_PROT::m_process_DISCONN_SYN_ACK_8(const void * buffer)
{
    const DISCONN_SYN_ACK_8 * disconnSynAck8 = static_cast<const DISCONN_SYN_ACK_8 *>(buffer);
    {
        boost::lock_guard<boost::mutex> lock(m_mutex);
        m_rnd = disconnSynAck8->rnd;
    }

#ifdef DEBUG
    std::cerr << ">> DISCONN_SYN_ACK (rnd: " << m_rnd << ")" << std::endl;
#endif

    m_netSend(DISCONN_ACK_N);

    if (!m_statusCb.empty())
        m_statusCb(false);

    {
        boost::lock_guard<boost::mutex> lock(m_mutex);
        m_phase = 5;
        m_phaseTime = GetTickCount();
    }
    m_cond.notify_one();
}
//---------------------------------------------------------------------------
void IA_CLIENT_PROT::m_process_FIN_8(const void *)
{
#ifdef DEBUG
    std::cerr << ">> FIN" << std::endl;
#endif

    {
        boost::lock_guard<boost::mutex> lock(m_mutex);
        m_phase = 1;
        m_phaseTime = GetTickCount();
    }
    m_cond.notify_one();
}
//---------------------------------------------------------------------------
void IA_CLIENT_PROT::m_process_INFO_8(const void * buffer)
{
    const INFO_8 * info = static_cast<const INFO_8 *>(buffer);

#ifdef DEBUG
    std::cerr << ">> INFO" << std::endl;
#endif
    std::string source(reinterpret_cast<const char *>(info->text));
    std::string dest;
    if (IconvString(source, dest, "CP1251", "UTF8")) {
        if (!m_messageCb.empty())
            m_messageCb(dest, info->infoType, info->showTime, info->sendTime);
#ifdef DEBUG
        std::cerr << "Message: '" << dest << "'" << std::endl;
#endif
    } else {
#ifdef DEBUG
        std::cerr << "Failed to convert '" << source << "' from CP1251 to UTF8" << std::endl;
#endif
    }
}
//---------------------------------------------------------------------------
void IA_CLIENT_PROT::m_process_ERROR(const void * buffer)
{
    const ERR_8 * err = static_cast<const ERR_8 *>(buffer);

#ifdef DEBUG
    std::cerr << ">> ERROR" << std::endl;
#endif

    std::string source(reinterpret_cast<const char *>(err->text));
    std::string dest;
    if (IconvString(source, dest, "KOI8-RU", "UTF8")) {
        if (!m_errorCb.empty())
            m_errorCb(dest, IA_SERVER_ERROR);
#ifdef DEBUG
        std::cerr << "Error: '" << dest << "'" << std::endl;
#endif
    } else {
#ifdef DEBUG
        std::cerr << "Failed to convert '" << source << "' from CP1251 to UTF8" << std::endl;
#endif
    }

    {
        boost::lock_guard<boost::mutex> lock(m_mutex);
        m_phase = 1;
        m_phaseTime = GetTickCount();
        m_codeError = IA_SERVER_ERROR;
        m_reconnect = false;
    }
    m_cond.notify_one();
}
//---------------------------------------------------------------------------
size_t IA_CLIENT_PROT::m_prepare_CONN_SYN_8(void * buffer, size_t size)
{
    assert(size >= sizeof(CONN_SYN_8));
    assert(sizeof(CONN_SYN_8) == Min8(sizeof(CONN_SYN_8)));

    CONN_SYN_8 * packet = static_cast<CONN_SYN_8 *>(buffer);

    packet->len = sizeof(CONN_SYN_8);
    strncpy(reinterpret_cast<char *>(packet->type), "CONN_SYN", IA_MAX_TYPE_LEN);
    strncpy(reinterpret_cast<char *>(packet->login), m_login.c_str(), IA_LOGIN_LEN);
    packet->dirs = 0;

    boost::lock_guard<boost::mutex> lock(m_mutex);
    for (size_t i = 0; i < DIR_NUM; i++) {
        packet->dirs |= (m_selectedDirs[i] << i);
    }

#ifdef DEBUG
    std::cerr << "<< CONN_SYN (login: '" << m_login << "', dirs: 0x" << std::hex << std::setw(8) << std::setfill('0') << packet->dirs << std::dec << ")" << std::endl;
#endif

#ifdef BOOST_BIG_ENDIAN
    SwapBytes(packet->len);
#endif

    return sizeof(*packet);
}
//---------------------------------------------------------------------------
size_t IA_CLIENT_PROT::m_prepare_CONN_ACK_8(void * buffer, size_t size)
{
    assert(size >= sizeof(CONN_ACK_8));
    assert(sizeof(CONN_ACK_8) == Min8(sizeof(CONN_ACK_8)));

    CONN_ACK_8 * packet = static_cast<CONN_ACK_8 *>(buffer);

    packet->len = sizeof(CONN_ACK_8);
    strncpy(reinterpret_cast<char *>(packet->type), "CONN_ACK", IA_MAX_TYPE_LEN);
    strncpy(reinterpret_cast<char *>(packet->loginS), m_login.c_str(), IA_LOGIN_LEN);

    boost::lock_guard<boost::mutex> lock(m_mutex);
    packet->rnd = ++m_rnd;

#ifdef DEBUG
    std::cerr << "<< CONN_ACK (login: '" << m_login << "', rnd: " << packet->rnd << ")" << std::endl;
#endif

#ifdef BOOST_BIG_ENDIAN
    SwapBytes(packet->len);
    SwapBytes(packet->rnd);
#endif

    return sizeof(*packet);
}
//---------------------------------------------------------------------------
size_t IA_CLIENT_PROT::m_prepare_ALIVE_ACK_8(void * buffer, size_t size)
{
    assert(size >= sizeof(ALIVE_ACK_8));
    assert(Min8(sizeof(ALIVE_ACK_8)) == sizeof(ALIVE_ACK_8));

    ALIVE_ACK_8 * packet = static_cast<ALIVE_ACK_8 *>(buffer);

    packet->len = sizeof(ALIVE_ACK_8);
    strncpy(reinterpret_cast<char *>(packet->type), "ALIVE_ACK", IA_MAX_TYPE_LEN);
    strncpy(reinterpret_cast<char *>(packet->loginS), m_login.c_str(), IA_LOGIN_LEN);

    boost::lock_guard<boost::mutex> lock(m_mutex);
    packet->rnd = ++m_rnd;

#ifdef DEBUG
    std::cerr << "<< ALIVE_ACK (login: '" << m_login << "', rnd: " << packet->rnd << ")" << std::endl;
#endif

#ifdef BOOST_BIG_ENDIAN
    SwapBytes(packet->len);
    SwapBytes(packet->rnd);
#endif

    return sizeof(*packet);
}
//---------------------------------------------------------------------------
size_t IA_CLIENT_PROT::m_prepare_DISCONN_SYN_8(void * buffer, size_t size)
{
    assert(size >= sizeof(DISCONN_SYN_8));
    assert(Min8(sizeof(DISCONN_SYN_8)) == sizeof(DISCONN_SYN_8));

    DISCONN_SYN_8 * packet = static_cast<DISCONN_SYN_8 *>(buffer);

    packet->len = sizeof(DISCONN_SYN_8);
    strncpy(reinterpret_cast<char *>(packet->type), "DISCONN_SYN", IA_MAX_TYPE_LEN);
    strncpy(reinterpret_cast<char *>(packet->loginS), m_login.c_str(), IA_LOGIN_LEN);
    strncpy(reinterpret_cast<char *>(packet->login), m_login.c_str(), IA_LOGIN_LEN);

#ifdef DEBUG
    std::cerr << "<< DISCONN_SYN (login: '" << m_login << "')" << std::endl;
#endif

#ifdef BOOST_BIG_ENDIAN
    SwapBytes(packet->len);
#endif

    return sizeof(*packet);
}
//---------------------------------------------------------------------------
size_t IA_CLIENT_PROT::m_prepare_DISCONN_ACK_8(void * buffer, size_t size)
{
    assert(size >= sizeof(DISCONN_ACK_8));
    assert(Min8(sizeof(DISCONN_ACK_8)) == sizeof(DISCONN_ACK_8));

    DISCONN_ACK_8 * packet = static_cast<DISCONN_ACK_8 *>(buffer);

    packet->len = Min8(sizeof(DISCONN_ACK_8));
    strncpy(reinterpret_cast<char *>(packet->loginS), m_login.c_str(), IA_LOGIN_LEN);
    strncpy(reinterpret_cast<char *>(packet->type), "DISCONN_ACK", IA_MAX_TYPE_LEN);

    boost::lock_guard<boost::mutex> lock(m_mutex);
    packet->rnd = ++m_rnd;

#ifdef DEBUG
    std::cerr << "<< DISCONN_ACK (login: '" << m_login << "', rnd: " << packet->rnd << ")" << std::endl;
#endif

#ifdef BOOST_BIG_ENDIAN
    SwapBytes(packet->len);
    SwapBytes(packet->rnd);
#endif

    return sizeof(*packet);
}
//---------------------------------------------------------------------------
int IA_CLIENT_PROT::m_determinatePacketType(const char * name) const
{
    const std::map<std::string, int>::const_iterator pi(m_packetTypes.find(name));
    if (pi == m_packetTypes.end()) {
        return -1;
    }
    return pi->second;
}
